/* avl.c : implementation details of an avl tree data structure for
   use with circlemud systems ONLY.
  
   This program can be used, redistributed or modified under the terms
   listed in the LICENSE file included with this package.  If you did
   not receive such a file, you should request one from the package
   provider.  

   This software is provided as is, without any warranty express or
   implied.  Use at your own risk.

   6/6/01 - Original revision - Patrick Dughi (dughi@imaxx.net)
*/

#include "conf.h"
#include "sysdep.h"

#include "structs.h"
#include "utils.h"
#include "interpreter.h"
#include "comm.h"

#include "avl.h"

struct avl_tree *avl_create(avl_compare *compare) {
  struct avl_tree *temp;

  if(compare == NULL) {
    return NULL;
  }

  temp = (struct avl_tree *) malloc(sizeof(struct avl_tree));  

  temp->avl = temp->retval = NULL;
  temp->compare = compare;
  temp->count = 0; 

  return temp;
}

int avl_insert(struct avl_tree *tree, void *value) {
  struct real_avl_tree **node=&(tree->avl);

  /* empty node, insert here */

  if(!tree->avl) {
    *node = (struct real_avl_tree *) malloc(sizeof(struct real_avl_tree));
    (*node)->left = NULL;
    (*node)->right = NULL;
    (*node)->balance = 0;
    (*node)->value = value;
    tree->count++;
    return 1;    
  }
  /* non-empty, perform node insert */
  return (avl_node_insert(tree->compare,node,value,&(tree->count),tree));
}

int avl_node_insert(avl_compare *compare, struct real_avl_tree **node,
                      void *value, int *count,struct avl_tree *tree) {
  int i;

  /* whoops! forgot this the first time through. couldn't figure it out.
	...oh yeah.. have to allocate the damned data! ..duh */

  if (!(*node)) {
    *node = malloc(sizeof(struct real_avl_tree));
    (*node)->left = (*node)->right = NULL;
    (*node)->value = value;
    (*count)++;
    (*node)->balance = 0;
    return 1;
  }

  if((i=compare((*node)->value, value)) > 0) {
    if(avl_node_insert(compare,&(*node)->right,value,count,tree) > 0) {
      return(avl_node_balance_right(node));
    }
  }
  if(i < 0) {
    if(avl_node_insert(compare,&(*node)->left,value,count,tree) > 0) {
      return (avl_node_balance_left(node));
    }

  }

	/* if insert is the same key, it's a replacement, and 
	   the original is returned in retval */

  if(!i) {
    tree->retval = ((*node)->value);
    (*node)->value = value;
    return 0;
  } else {
    return 1;
  }
  /* add error here! */
  return -1;
}

int avl_node_balance_right(struct real_avl_tree **node) {
  if((*node)->balance == -1) {
    (*node)->balance = 0;
    return 0;
  } else if((*node)->balance == 0) {
    (*node)->balance = 1;
    return 1;
  } else if((*node)->balance == 1) {   /* bleh. hard one */
    if((*node)->right->balance == 1) {
      (*node)->balance = (*node)->right->balance = 0;
      avl_rl(node);
    } else {
      if((*node)->right->left->balance == -1) {
        (*node)->balance = 0;
        (*node)->right->balance = 1;
      } else if((*node)->right->left->balance == 0) {
        (*node)->balance = (*node)->right->balance =0;
      } else if((*node)->right->left->balance == -1) {
        (*node)->balance = -1;
        (*node)->right->balance = 0;
      } else {
  	/* error here */
        return -1;
      }
      (*node)->right->left->balance = 0;
      avl_rr(&((*node)->right));
      avl_rl(node);
    }
    return 0;
  }
  /* add error here too! */
  return -1;
}

int avl_node_balance_left(struct real_avl_tree **node) {
  if((*node)->balance == -1) {
    if((*node)->left->balance == -1) {
      (*node)->balance = (*node)->left->balance = 0;
      avl_rr(node);
    } else {
      if((*node)->left->right->balance == -1) {
        (*node)->balance = 1;
        (*node)->right->balance = 0;
      } else if((*node)->left->right->balance == 0) {
        (*node)->balance = (*node)->left->balance =0;
      } else if((*node)->left->right->balance == -1) {
        (*node)->balance = 0;
        (*node)->left->balance = -1;
      } else {
  	/* error here */
        return -1;
      }
      (*node)->left->right->balance = 0;
      avl_rl(&((*node)->left));
      avl_rr(node);
    }
    return 0;
  } else if((*node)->balance == 0) {
    (*node)->balance = -1;
    return 1;
  } else if((*node)->balance == 1) { 
    (*node)->balance = 0;
    return 0;
  }
  /* add error here too! */
  return -1;
}

void avl_rr(struct real_avl_tree **node) {
  struct real_avl_tree *placeholder = *node;

  (*node) = (*node)->left;
  placeholder->left = (*node)->right;
  (*node)->right = placeholder;
}

void avl_rl(struct real_avl_tree **node) {
  struct real_avl_tree *placeholder = *node;
  
  (*node) = (*node)->right;
  placeholder->right = (*node)->left;
  (*node)->left = placeholder;
}

int avl_delete(struct avl_tree *tree, void *value) {
  struct real_avl_tree **node=&(tree->avl);

  if(!(*node)) {
    return -1;
  }
  
  /* otherwise do a node delete */
  return( avl_node_delete(tree->compare,node,value,&(tree->count),tree));
}

int avl_node_delete(avl_compare *compare, struct real_avl_tree **node,
	void *value, int *count,struct avl_tree *tree) {
  int result = 1;

  if(!*node) {
    tree->retval = NULL;
    return;
  }

  if(compare((*node)->value, value) > 0) {
    if((result = avl_node_delete(compare,&((*node)->right),value,count,tree)) == 1) {
      return (avl_b_rr(node));
    }   
    return result;
  } else if(compare((*node)->value, value) < 0) {
    if((result=avl_node_delete(compare,&((*node)->left),value,count,tree)) == 1) {
      return (avl_b_rl(node));
    }   
    return result;
  }

  if((*node)->right) {
    if(avl_rightmost((*node),&((*node)->right),&result)) {
      if(result == 1) {
        result = avl_b_rr(node);
      }
      return result;
    }
  }

  if((*node)->left) {
    if(avl_leftmost((*node),&((*node)->right),&result)) {
      if(result == 1) {
        result = avl_b_rl(node);
      }
      return result;
    }
  }

  /* found something */
  tree->retval = (*node)->value;
  free(*node);
  *node = NULL;
  (*count)--;
  /* have to fix it now */
  return 1;

}
  
int avl_b_rl(struct real_avl_tree **n) {
  if((*n)->balance == 1) {
    (*n)->balance = 0;
    return 1;
  } else if((*n)->balance == -1) {
    if ((*n)->left->balance == -1) {
      (*n)->balance = (*n)->left->balance = 0;
      avl_rr(n);
      return 1;
    } else if ((*n)->left->balance == 0) {
      (*n)->balance = -1;
      (*n)->left->balance = 1;
      avl_rr(n);
      return 0;
    } else {
      if ((*n)->left->right->balance == -1) {
	(*n)->balance = 1;
	(*n)->left->balance = 0;
      } else if ((*n)->left->right->balance == 1) {
	(*n)->balance = 0;
	(*n)->left->balance = -1;	
      } else {
	(*n)->balance = 0;
	(*n)->left->balance = 0;
      }
      (*n)->left->right->balance = 0;
      avl_rl(& (*n)->left);
      avl_rr(n);
      return 1;
    }
  } else {
    (*n)->balance = -1;
    return 0;
  }
}

int avl_b_rr(struct real_avl_tree **n) {
  if((*n)->balance == 1) {
    (*n)->balance = 0;
    return 1;
  } else if((*n)->balance == -1) {
    if ((*n)->left->balance == -1) {
      (*n)->balance = (*n)->left->balance = 0;
      avl_rr(n);
      return 1;
    } else if ((*n)->left->balance == 0) {
      (*n)->balance = -1;
      (*n)->left->balance = 1;
      avl_rr(n);
      return 0;
    } else {
      if ((*n)->left->right->balance == -1) {
	(*n)->balance = 1;
	(*n)->left->balance = 0;
      } else if ((*n)->left->right->balance == 1) {
	(*n)->balance = 0;
	(*n)->left->balance = -1;	
      } else {
	(*n)->balance = 0;
	(*n)->left->balance = 0;
      }
      (*n)->left->right->balance = 0;
      avl_rl(& (*n)->left);
      avl_rr(n);
      return 1;
    }
  } else {
    (*n)->balance = -1;
    return 0;
  }
}

int avl_rightmost(struct real_avl_tree *target, struct real_avl_tree **n,
				int *retval)
{
	struct real_avl_tree *tmp;

	*retval = 1;
	if (!(*n)) {
		return 0;
	}
	if ((*n)->right) {
		if (!avl_rightmost(target, &(*n)->right, retval)) {
			return 0;
		}
		if (*retval == 1) {
			*retval = avl_b_rr(n);
		}
		return 1;
	}
	target->value  = (*n)->value;
	tmp = *n;
	*n = (*n)->left;

	free(tmp);
	return 1;
}

int avl_leftmost(struct real_avl_tree *target, struct real_avl_tree **n,
		int *retval)
{
	struct real_avl_tree *tmp;

	*retval = 1;
	if (!(*n)) {
		return 0;
	}
	if ((*n)->left) {
		if (!avl_leftmost(target, &(*n)->left, retval)) {
			return 0;
		}
		if (*retval == 1) {
			*retval =  avl_b_rl(n);
		}
		return 1;
	}
	target->value = (*n)->value;
	tmp = *n;
	*n = (*n)->right;

	free(tmp);
	return 1;
}

void *avl_search(struct avl_tree *tree, void *value) {
  struct real_avl_tree *node = tree->avl;
  int i;

  if(!node) {
    return NULL;
  }
  if((i=tree->compare((node)->value, value)) > 0) {
    return avl_node_search(tree->compare,node->right,value);
  }
  if(i < 0) {
    return avl_node_search(tree->compare,node->left,value);
  }

  return (node->value);

}

void *avl_node_search(avl_compare *compare, struct real_avl_tree * node, 
		void *value) {
  int i;

  if(!node) {
    return NULL;
  }
  if((i=compare((node)->value, value)) > 0) {
    return avl_node_search(compare,node->right,value);
  }
  if(i < 0) {
    return avl_node_search(compare,node->left,value);
  }

  return (node->value);
  
}

int avl_count(struct avl_tree *tree) {
  return (tree->count);
}

/* hm. have to put this in eventually, i guess */
void avl_delete_all(struct avl_tree *tree, avl_clear *del) {
  avl_delete_all_node(tree->avl,del);
  if(tree) {
    free(tree);
    tree=NULL;
  } 
}

void avl_delete_all_node(struct real_avl_tree *node, avl_clear *del)
{
  /* do left, then right, then me */
  /* technically we should never return here ... */
  if(!node) {
    return;
  }
 
  if(node->left) {
    avl_delete_all_node(node->left,del);
  } 
  if(node->right) {
    avl_delete_all_node(node->right,del);
  }
  del(node->value);


  free(node);
  node = NULL;
  return;
}

/* the logic for avl tree traversal is simple;

init and store:
   upon initilization, make a linked list of pointers to the nodes to the
   left.  Eventually, the list will terminate (can no longer go left).
  	
operate:
   Immediately after you pop a node from the list, add to the list
   the tree to the right of the just-popped node pointer, using the same
   alogrithm as above (only left).  By following this pattern, you will
   have a linked list at _most_ N nodes deep, where N is the depth of the
   avl tree


example:

    tree:     g		 initial list = a - b - g
	     / \	    pop "a"
            b   u           b-g
           / \   \          pop "b"  add c
          a   c   z         c-g
	       \            pop "c" add d
	        d           d-g
                            pop "d"
                            pop "g" add u
			    u
			    pop "u" add z
			    z
			    pop "z" 

	Look at that - saves state and very very small memory, and even
	though it's in a linked list, it's incredibly short, and you only
        pop from the top of the list, you don't have to worry about
	searching or lookup times, or anything.  Pretty sweet.
*/
	                           

struct avl_tree_traverse *init_traverse_tree(struct avl_tree *tree) {
  struct avl_tree_traverse *t=NULL;

  if(!tree || !tree->avl) {
    return NULL;
  }
  
  /* now, add to left */
  init_traverse_node(tree->avl,&t);

  return t;

}

void init_traverse_node(struct real_avl_tree *node,
		struct avl_tree_traverse **t) {
  struct avl_tree_traverse *temp;

  temp = (struct avl_tree_traverse *) malloc(sizeof(struct avl_tree_traverse));
  temp->node = node;
  temp->next = *t;
  *t = temp;
  
  /* left */
  if(node->left) {
    init_traverse_node(node->left,t);
  }
  
}

  /* here's the trick, pop the top node, and add the tree to the
     uh.. right. */

void *avl_traverse(struct avl_tree_traverse **t) {
  struct real_avl_tree *temp;
  struct avl_tree_traverse *n;

  if(t && (*t)) {
    n = (*t)->next; 
    temp = (*t)->node;

    free(*t);
    *t = n;
    if(temp->right) {
      init_traverse_node(temp->right,t);
    }
    return (temp->value);
  }

  return NULL;   
}

void avl_traverse_free(struct avl_tree_traverse *t) {
  struct avl_tree_traverse *temp;

  for(;t;t=temp) {
    temp = t->next;
    free(t);
  }
}

/* these two functions don't actually _DO_ anything */
/* They are examples of how to traverse the whole tree inorder */

void avl_traverse_all_nodes(struct real_avl_tree *t) {
  if(t) {
    avl_traverse_all_nodes(t->left);
    avl_traverse_all_nodes(t->right);
  } 
}

void avl_traverse_all(struct avl_tree *t) {
  avl_traverse_all_nodes(t->avl);
}

void number_test(char *argument);
void room_test(char *argument);
void string_test(char *argument);

ACMD(do_avltest) {
  char buf[400];
  char *crud;
  char *holder;
  if(!*argument) {
    send_to_char("Choices are\n"
 	 	 "Number test: avltest 1 [number between 1-10 to delete]\n"
                 "Room test  : avltest 2 [number between 1-10 to delete]\n"
		 "String test: avltest 3 [string 'zero' through 'nine'\n",ch);
    return;
  }  

  holder = crud=strdup(argument);

  crud=one_argument(crud,buf);

  switch(atoi(buf)) {
    case 1:
      number_test(crud);
      break;
    case 2:
      room_test(crud);      
      break;
    case 3:
      string_test(crud);
      break;
    default:
      send_to_char("That is not a choice.\n",ch);
      break;
  }

  free(holder);
  return;
}


/*
   this is our generic delete function.  This will work on any data
   structures which are simple enough to have not required any additional
   memory allocations.  Observant viewers may realize this function could
   have been replaced by just giving the delete_all function a pointer to
   the free command.  But i like being neat and setting pointers to NULL. 

   Besides, I'm writing this as an example.

   More complex deletion functions will be required for your more complex
   data structures (like rooms or chars that have strings set).

   In many cases, you probably already have a deletion function for those
   types; in this case, just write a wrapper function to cast the data
   from a void * to a your_Data_type_here *.
*/

void generic_delete(void *our_simple_data) {
  if(our_simple_data) {
    free(our_simple_data);
    our_simple_data = NULL;
  }
}

int compare_numbers(void *left, void *right) {

  if(*((int *)left) > *((int *)right)) {
    return 1;
  }
  if(*((int *)left) < *((int *)right)) {
    return -1;
  }
  return 0;

}

void number_test(char *argument) {
  struct avl_tree *heck;
  struct avl_tree_traverse *trav;
  int *temp=NULL;

  int *x[10];
  int i;
  char buf[400];

  for(i=0;i<10;i++) {
    x[i] = (int *) malloc(sizeof(int));
  }
  *x[0] = 5;
  *x[1] = 4;
  *x[2] = 2;
  *x[3] = 3;
  *x[4] = 0;
  *x[5] = 1;
  *x[6] = 7;
  *x[7] = 9;
  *x[8] = 6;
  *x[9] = 8;

  heck = avl_create(compare_numbers);

  for(i=0;i<10;i++) {
    avl_insert(heck,(void *) x[i]);
  }
  one_argument(argument,buf);
  if(*buf) {
    i = atoi(buf);
    if(i>=0 && i < 10) {
      avl_delete(heck, &i); 
      if((temp = (int *) avl_search(heck, &i))) {
        log("Still there, after deletion!");
      } else {
        log("successfully deleted '%d'",i);
        /* better free it */

        free(heck->retval);
        heck->retval = NULL;
      }
    } else {
      log("Could not find '%d'",i);
    }
  }

  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("null trav");
  }
  while(trav) {
    i = * ((int *)avl_traverse(&trav));
    log("Traversal: '%d'",i);
  }

  log("Got '%d' for count",avl_count(heck));

  log("Attempting to add a duplicate record");

  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("Null trav");
  }
  
  i = *((int *)avl_traverse(&trav));
  log("Got '%d'",i);

  /* not going to use traverse anymore */
  avl_traverse_free(trav);

  temp = (int *) malloc (sizeof(int));
  *temp = i;
  avl_insert(heck,(void *) temp);
   temp = (int *) heck->retval;
  if(temp) {
    log("Got '%d' for return value",*temp);
  }

  free(heck->retval);
  heck->retval = NULL;  

  /* need to free everthing up */
  avl_delete_all(heck, generic_delete);
}

int compare_rooms(void *left, void *right) {

  if(((struct room_data *)left)->number > 
     ((struct room_data *)right)->number ) {
    return 1;
  }

  if(((struct room_data *)left)->number <
     ((struct room_data *)right)->number ) {
    return -1;
  }
  return 0;
}

void room_test(char *argument) {
  struct avl_tree *heck;
  struct avl_tree_traverse *trav;
  struct room_data *temp=NULL;
  int i;
  struct room_data *x[10];
  char buf[400];

  for(i=0;i<10;i++) {
    x[i] = (struct room_data *) malloc(sizeof(struct room_data));
  }

  x[0]->number = 5;
  x[1]->number = 4;
  x[2]->number = 2;
  x[3]->number = 3;
  x[4]->number = 0;
  x[5]->number = 1;
  x[6]->number = 7;
  x[7]->number = 9;
  x[8]->number = 6;
  x[9]->number = 8;
  
  /* for duplication checking, later */

  x[0]->sector_type = 88;
  x[1]->sector_type = 88;
  x[2]->sector_type = 88;
  x[3]->sector_type = 88;
  x[4]->sector_type = 88;
  x[5]->sector_type = 88;
  x[6]->sector_type = 88;
  x[7]->sector_type = 88;
  x[8]->sector_type = 88;
  x[9]->sector_type = 88;

  heck = avl_create(compare_rooms);

  for(i=0;i<10;i++) {
    avl_insert(heck,(void *) x[i]);
  }
  one_argument(argument,buf);
  if(*buf) {
    temp = (struct room_data *) malloc(sizeof(struct room_data));
    temp->number = atoi(buf);
    if(temp->number >= 0 && temp->number < 10) {
      avl_delete(heck, temp); 
      if(avl_search(heck, temp)) {
        log("Still there, after deletion!");
      } else {
        log("successfully deleted room '%d'",temp->number);
      }
    } else {
      log("Could not find room '%d'",temp->number);
    }

    free(temp);
    temp=NULL;
    if(heck->retval) {

      free(heck->retval);
      heck->retval=NULL;
    }
  }
  
  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("null trav");
  }
  while(trav) {
    temp =  ((struct room_data *)avl_traverse(&trav));
    log("Traversal of room: '%d'",temp->number);
  }

  log("Got '%d' for count",avl_count(heck));

  log("Attempting to add a duplicate record");

  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("Null trav");
  }
  
  temp = (struct room_data *) malloc(sizeof(struct room_data));
  temp->number = ((struct room_data *)avl_traverse(&trav))->number;
  temp->sector_type = 77;
  log("Got room '%d'",temp->number);

  /* not going to use traverse anymore */
  avl_traverse_free(trav);

  x[0] = temp;
  avl_insert(heck,(void *) x[0]);

  /* note; we're guarenteed that after we insert that THAT particular
     piece of memory is saved in the tree, so we're free to reassign
     the pointer (and it's probably a good idea to do that anyway) */

  temp = (struct room_data *) heck->retval;
  
  if(temp) {
    log("Got room '%d, sector_type '%d' for return value",temp->number, temp->sector_type);
    /* just for fun, lets look and see what's in there now! */
    if((temp=avl_search(heck, heck->retval))) {
      log("Current room '%d', has sector_type '%d'",temp->number,temp->sector_type);
    }
  }

  free(heck->retval);
  heck->retval = NULL;  

  /* need to free everthing up */
  /* we can use our generic delete because we didn't allocate any memory
     to pointers/strings/etc */
  avl_delete_all(heck, generic_delete);  

}

int compare_strings(void *left, void *right) {

  return strcmp((char *)left,(char *)right);

}

void string_test(char *argument) {
  struct avl_tree *heck;
  struct avl_tree_traverse *trav;
  char *temp=NULL;
  int i;
  char *x[10];
  char buf[400];

  x[0] = strdup("zero");
  x[1] = strdup("one");
  x[2] = strdup("two");
  x[3] = strdup("three");
  x[4] = strdup("four");
  x[5] = strdup("five");
  x[6] = strdup("six");
  x[7] = strdup("seven");
  x[8] = strdup("eight");
  x[9] = strdup("nine");

  heck = avl_create(compare_strings);

  for(i=0;i<10;i++) {
    avl_insert(heck,(void *) x[i]);
  }
  one_argument(argument,buf);
  if(*buf) {
    temp = strdup(buf);
    avl_delete(heck, temp); 
    if(heck->retval) {
      if(avl_search(heck, temp)) {
        log("Still there, after deletion!");
      } else {
        log("successfully deleted '%s'",temp);
      }
    } else {
      log("That value wasn't found");
    }
    free(temp);
    temp=NULL;
    if(heck->retval) {
      free(heck->retval);
      heck->retval=NULL;
    }
  }
  
  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("null trav");
  }
  while(trav) {
    temp =  ((char *)avl_traverse(&trav));
    log("Traversal garnered string: '%s'",temp);
  }

  log("Got '%d' for count",avl_count(heck));

  log("Attempting to add a duplicate record");

  trav = init_traverse_tree(heck);

  if(trav == NULL) {
    log("Null trav");
  }
  
  temp = strdup((char *)avl_traverse(&trav));
  log("Got string '%s' to replace with identical",temp);

  /* not going to use traverse anymore */
  avl_traverse_free(trav);

  avl_insert(heck,(void *) temp);

  /* note; we're guarenteed that after we insert that THAT particular
     piece of memory is saved in the tree, so we're free to reassign
     the pointer (and it's probably a good idea to do that anyway) */
  
  temp = (char *) heck->retval;
  
  if(temp) {
     log("Got the string '%s' back from insertion",temp);
  }
  free(heck->retval);
  heck->retval = NULL;  

  /* need to free everthing up */
  /* we can use our generic delete because we didn't allocate any memory
     to pointers/strings/etc */
  avl_delete_all(heck, generic_delete);  

}

